#include <algorithm>
#include <stdexcept>
#include <cassert>
#include <typeinfo>
#include <cmath>

#include "sgui.h"





#ifdef _WIN32
  #include <windows.h>
#else
  #include <dlfcn.h>
  #include <unistd.h>
  #include <sys/stat.h>
  #include <sys/types.h>
  #include <dirent.h>
#endif
#include <SDL/SDL_opengl.h>

namespace sgui {



class MatrixSaver
{
public:

// GL_MODELVIEW_MATRIX | GL_PROJECTION_MATRIX | GL_TEXTURE_MATRIX
  MatrixSaver(int matrix_type = GL_MODELVIEW_MATRIX) {
    glGetFloatv(matrix_type, m_matrix);

    switch(matrix_type) {
    case GL_MODELVIEW_MATRIX:
      m_matrix_mode = GL_MODELVIEW;
      break;
    case GL_PROJECTION_MATRIX:
      m_matrix_mode = GL_PROJECTION;
      break;
    case GL_TEXTURE_MATRIX:
      m_matrix_mode = GL_TEXTURE;
      break;
    default:
      m_matrix_mode = GL_MODELVIEW;
      break;
    }
  }

  ~MatrixSaver() {
    int current_matrix_mode;
    glGetIntegerv(GL_MATRIX_MODE, &current_matrix_mode);

    glMatrixMode(m_matrix_mode);
    glLoadMatrixf(m_matrix);
    glMatrixMode(current_matrix_mode);
  }

private:
  int m_matrix_mode;
  float m_matrix[16];
};





void AssignColor(const Color& c)
{
  glColor4fv(c.v);
}

void AssignColorO(const Color& c, float opa)
{
  Color a = c;
  a.a*=opa;
  glColor4fv(a.v);
}

void DrawPoint(const Point& p) {
  glBegin(GL_POINTS);
  glVertex2f(p.getX(), p.getY());
  glEnd();
}

void DrawLine(const Point& p1, const Point& p2)
{
  glBegin(GL_LINES);
  glVertex2f(p1.getX(), p1.getY());
  glVertex2f(p2.getX(), p2.getY());
  glEnd();
}

void DrawRectEdge(const Rect& r)
{
  glBegin(GL_LINE_LOOP);
  glVertex2f(r.getLeft(), r.getTop());
  glVertex2f(r.getRight(), r.getTop());
  glVertex2f(r.getRight(), r.getBottom());
  glVertex2f(r.getLeft(), r.getBottom());
  glEnd();
}

void DrawRect(const Rect& r)
{
  glBegin(GL_QUADS);
  glTexCoord2f(0.0f, 0.0f);
  glVertex2f(r.getLeft(), r.getBottom());
  glTexCoord2f(1.0f, 0.0f);
  glVertex2f(r.getRight(), r.getBottom());
  glTexCoord2f(1.0f, 1.0f);
  glVertex2f(r.getRight(), r.getTop());
  glTexCoord2f(0.0f, 1.0f);
  glVertex2f(r.getLeft(), r.getTop());
  glEnd();
}




FontRenderer::FontRenderer() : m_font(0), m_halign(FontRenderer::HLEFT), m_valign(FontRenderer::VTOP)
{
  m_font = App::instance()->getDefaultFont();
}

void FontRenderer::setHAlign(HAlign align) { m_halign = align; }
void FontRenderer::setVAlign(VAlign align) { m_valign = align; }
void FontRenderer::setFont(Font *font) { m_font = font; }

void FontRenderer::draw(const wstring& str, const Point& pos, const Rect& _edge) const
{
  if(!m_font || str.empty()) {
    return;
  }

  Rect edge = _edge;
  if(edge.getWidth()==0.0f) {
    edge = Rect(getAdvance(str), 0.0f, getHeight(), 0.0f);
  }
  MatrixSaver ms;

  Point rel = pos;
  rel.setY(rel.getY()+m_font->LineHeight());
  if(m_valign==FontRenderer::VCENTER) {
    rel.setY(rel.getY()+(edge.getHeight()-getHeight(str))*0.5f-1.0f);
  }
  else if(m_valign==FontRenderer::VBOTTOM) {
    rel.setY(edge.getBottom()-getHeight(str));
  }

  if(m_halign==FontRenderer::HCENTER) {
    rel.setX(rel.getX()+(edge.getWidth()-getAdvance(str))*0.5f);
  }
  else if(m_halign==FontRenderer::HRIGHT) {
    rel.setX(edge.getRight()-getAdvance(str));
  }

  glEnable(GL_TEXTURE_2D);
  glTranslatef(rel.getX(), rel.getY(), 0.0f);
  glScalef(1.0f, -1.0f, 1.0f);

  static wstring buf;
  size_t i = 0;
  for(;;) {
    float width = 0.0f;
    for(;;) {
      wchar_t c = str[i];
      if(c=='\n' || c==0) {
        ++i;
        break;
      }
      else {
        float adv = getAdvance(c);
        width+=adv;
        buf+=c;
        ++i;
        if(width>edge.getWidth()) {
          break;
        }
      }
    }
    glPushMatrix();
    m_font->Render(buf.c_str());
    glPopMatrix();
    buf.clear();

    if(i>=str.size()) {
      break;
    }
    else {
      glTranslatef(0, -(getHeight()+1.0f), 0.0f);
    }
  }
  glBindTexture(GL_TEXTURE_2D,0);
  glDisable(GL_TEXTURE_2D);
}

Font* FontRenderer::getFont() const { return m_font; }
float FontRenderer::getHeight() const
{
  if(!m_font) {
    return 0;
  }

  return std::ceil(m_font->LineHeight());
}

float FontRenderer::getHeight(const wstring& str) const
{
  int l = 1;
  for(size_t i=0; i<str.size(); ++i) {
    if(str[i]=='\n') {
      ++l;
    }
  }
  return getHeight()*l;
}

float FontRenderer::getAdvance(wchar_t c) const
{
  if(!m_font) {
    return 0;
  }

  wchar_t buf[2] = {0, 0};
  buf[0] = c;
  return m_font->Advance(buf);
}

float FontRenderer::getAdvance(const wchar_t *str, size_t len) const
{
  if(!m_font) {
    return 0;
  }

  static std::vector<wchar_t> buf;
  buf.assign(str, str+len);
  buf.push_back(0);
  return m_font->Advance(&buf[0]);
}

float FontRenderer::getAdvance(const wstring& str) const
{
  if(!m_font) {
    return 0;
  }

  float width = 0.0f;
  std::vector<wchar_t> buf;
  for(size_t i=0; i<str.size(); ++i) {
    wchar_t c = str[i];
    if(c=='\n' || i==str.size()-1) {
      if(i==str.size()-1) {
        buf.push_back(c);
      }
      buf.push_back(0);
      width = std::max<float>(width, m_font->Advance(&buf[0]));
      buf.clear();
    }
    else {
      buf.push_back(c);
    }
  }
  return width;
}



StopWatch::StopWatch() :
  m_start_time(0), m_end_time(0)
{}

void StopWatch::run(size_t limit)
{
  m_start_time = SDL_GetTicks();
  m_end_time = m_start_time+limit;
}

void StopWatch::reset()
{
  m_start_time = 0;
  m_end_time = 0;
}

void StopWatch::stop()
{
  m_end_time = SDL_GetTicks();
}

bool StopWatch::isRunning() const
{
  size_t t = SDL_GetTicks();
  return t>=m_start_time && t < m_end_time;
}

bool StopWatch::isFinished() const
{
  return m_start_time!=0 && SDL_GetTicks()>=m_end_time;
}

bool StopWatch::isActive() const
{
  return isRunning() || isFinished();
}

size_t StopWatch::getLeft() const
{
  size_t t = SDL_GetTicks();
  if(m_end_time > t) {
    return m_end_time-t;
  }
  return 0;
}

size_t StopWatch::getPast() const
{
  return std::min<size_t>(SDL_GetTicks()-m_start_time, m_end_time-m_start_time);
}


Uint32 Timer::callback(Uint32 interval, void *param) {
  ((Timer*)param)->onTimer(interval);
  return interval;
}

Timer::Timer() :
  m_timerid(0), m_target(0), m_autorestart(0)
{
}

Timer::~Timer() {
  stop();
}

void Timer::setTarget(Window *target) { m_target = target; }

void Timer::run(size_t interval, bool autorestart) {
  if(m_timerid) {
    stop();
  }
  m_autorestart = autorestart;
  m_timerid = SDL_AddTimer(interval, &callback, this);
}

void Timer::stop() {
  if(m_timerid) {
    SDL_RemoveTimer(m_timerid);
    m_timerid = 0;
  }
}

void Timer::onTimer(size_t interval) {
  App::instance()->getEventServer().queueEvent(new TimerEvent(EVT_TIMER, 0, m_target));
  if(!m_autorestart) {
    stop();
  }
}







//  RootObject

Object::Object() {}
Object::~Object() {}
void Object::release() {
//  printf("sgui::Object::~Object() : %s\n", typeid(*this).name());
  delete this;
}

#ifdef SG_BUILD_DSO
void* Object::operator new(size_t size) { return malloc(size); }
void Object::operator delete(void *p) { return free(p); }
#endif // SG_BUILD_DSO



//  Event

Event::Event(EventType type, Window *src, Window *dst) :
  m_type(type), m_src(src), m_dst(dst), m_hcontrol(0), m_hframe(0)
{
}

EventType Event::getType(void) const { return m_type; }
Window* Event::getSrc() const { return m_src; }
Window* Event::getDst() const { return m_dst; }
int Event::isHandled() const { return m_hcontrol+m_hframe; }
int Event::isHandledByControl() const { return m_hcontrol; }
int Event::isHandledByFrame() const { return m_hframe; }


DestructEvent::DestructEvent(EventType type, Window *src, Window *dst, Window *self) :
  Event(type, src, dst), m_self(self), m_selfid(self->getID())
{
}

KeyboardEvent::KeyboardEvent(EventType type, Window *src, Window *dst,
  unsigned char scan_code, SDLMod modifiers, SDLKey key, Uint16 unicode) :
  Event(type, src, dst), m_scan_code(scan_code), m_modifiers(modifiers), m_key(key), m_unicode(unicode)
{
}

MouseEvent::MouseEvent(EventType type, Window *src, Window *dst,
  Point point, Point relative, unsigned int button) :
  Event(type, src, dst), m_real_position(point), m_real_relative(relative), m_button(button)
{
  m_position = View::instance()->toOpenGLCoord(m_real_position);
  m_relative = View::instance()->toOpenGLCoord(m_real_relative);
}


unsigned int MouseEvent::TranslateSDLButton(Uint8 SDLButton) {
  unsigned int Button = 0;
  switch (SDLButton) {
  case SDL_BUTTON_LEFT: Button = LEFT; break;
  case SDL_BUTTON_RIGHT: Button = RIGHT; break;
  case SDL_BUTTON_MIDDLE: Button = MIDDLE; break;
  case SDL_BUTTON_WHEELUP: Button = WHEELUP; break;
  case SDL_BUTTON_WHEELDOWN: Button = WHEELDOWN;  break;
  default: break;
  }
  return Button;
}

unsigned int MouseEvent::TranslateSDLButtonState(Uint8 SDLButtonState) {
  unsigned int Button = 0;
  if(SDLButtonState & SDL_BUTTON(1)) Button |= LEFT;
  if(SDLButtonState & SDL_BUTTON(2)) Button |= MIDDLE;
  if(SDLButtonState & SDL_BUTTON(3)) Button |= RIGHT;
  if(SDLButtonState & SDL_BUTTON(4)) Button |= WHEELUP;
  if(SDLButtonState & SDL_BUTTON(5)) Button |= WHEELDOWN;
  return Button;
}


JoyEvent::JoyEvent(EventType type, Window *src, Window *dst,
  int devid, int button, int x, int y) :
  Event(type, src, dst), m_devid(devid), m_button(button), m_axis_x(x), m_axis_y(y)
{
}







//  EventServer

EventServer::EventServer() {}
EventServer::~EventServer() { clearEvent(); }

void EventServer::registerEventClient(EventClient *client, EventType type) {
  if(!client) {
    return;
  }

  eventclient_cont& cont = m_evtclients[type];
  eventclient_cont::iterator p = std::find(cont.begin(), cont.end(), client);
  if(p==cont.end()) {
    cont.push_front(client);
  }
}

void EventServer::deregisterEventClient(EventClient *client, EventType type) {
  eventclient_cont& cont = m_evtclients[type];
  cont.remove(client);
}

void EventServer::deregisterEventClient(EventClient *client) {
  for(eventclient_map::iterator p=m_evtclients.begin(); p!=m_evtclients.end(); ++p) {
    p->second.remove(client);
  }
  m_toplevel.remove(client);

  for(event_cont::iterator p=m_events.begin(); p!=m_events.end(); /**/) {
    Event *e = *p;
    if(e->getSrc()==client || e->getDst()==client) {
      m_events.erase(p++);
      e->release();
    }
    else {
      ++p;
    }
  }
}

void EventServer::setTopLevel(EventClient *client)
{
  m_toplevel.remove(client);
  m_toplevel.push_back(client);
}

void EventServer::queueEvent(Event *evt)
{
  ScopedLock lock(m_mutex);

  if(!!evt) {
    m_events.push_back(evt);
  }
}

void EventServer::deliverEvent()
{
  ScopedLock lock(m_mutex);

  while(!m_events.empty()) {
    Event *e = m_events.front();
    m_events.pop_front();
    eventclient_cont& cli = m_evtclients[e->getType()];

    for(eventclient_cont::iterator q=cli.begin(); q!=cli.end(); ++q) {
      EventClient *c = *q;
      if(c->handleEvent(*e)) {
        if(Panel *panel = dynamic_cast<Panel*>(c)) {
          e->m_hframe++;
        }
        else {
          e->m_hcontrol++;
        }
      }
    }

    for(eventclient_cont::iterator q=m_toplevel.begin(); q!=m_toplevel.end(); ++q) {
      for(eventclient_map::iterator p=m_evtclients.begin(); p!=m_evtclients.end(); ++p) {
        eventclient_cont& cont = p->second;
        cont.remove(*q);
        cont.push_front(*q);
      }
    }
    m_toplevel.clear();

    App::instance()->handleEvent(*e);

    e->release();
  }
}

void EventServer::clearEvent()
{
  ScopedLock lock(m_mutex);

  while(!m_events.empty()) {
    Event *e = m_events.front();
    m_events.pop_front();
    e->release();
  }
}



//  EventClient

EventClient::EventClient()
{}

EventClient::~EventClient()
{
  App::instance()->getEventServer().deregisterEventClient(this);
}

void EventClient::listenEvent(EventType type)
{
  App::instance()->getEventServer().registerEventClient(this, type);
}

void EventClient::queueEvent(Event *e)
{
  App::instance()->getEventServer().queueEvent(e);
}

bool EventClient::handleEvent(const Event& e)
{
  return false;
}


//  Window

Window::Window(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  m_id(id), m_visible(true), m_floating(false), m_parent(0)
{
  setParent(parent);
  setPosition(pos);
  setSize(size);
  setText(text);

  setForegroundColor(App::instance()->getDefaultForegroundColor());
  setBackgroundColor(App::instance()->getDefaultBackgroundColor());
  setBorderColor(App::instance()->getDefaultBorderColor());

  queueEvent(new Event(EVT_CONSTRUCT, this, 0));
}

Window::~Window()
{
  releaseChildren();
  queueEvent(new DestructEvent(EVT_DESTRUCT, 0, 0, this));
  setParent(0);
}

void Window::releaseChildren()
{
  while(!m_children.empty()) {
    (*m_children.begin())->release();
  }
}

bool Window::isInside(const Point& point) const
{
  Rect rect(getGlobalPosition(), getSize());
  if(!isFloating() && !!getParent()) {
    getParent()->clip(rect, this);
  }
  return rect.isInside(point);
}

Point Window::toGlobalCoord(const Point& local) const
{
  Point global = local;
  global+=getClientPosition();
  global+=getPosition();
  if(!!getParent()) {
    global = getParent()->toGlobalCoord(global);
  }
  return global;
}

Point Window::toLocalCoord(const Point& global) const
{
  Point local = global;
  local-=getClientPosition();
  local-=getPosition();
  if(!!getParent()) {
    local = getParent()->toLocalCoord(local);
  }
  return local;
}

void Window::clip(Rect& r, const Window *wnd) const
{
  if(!!wnd && wnd->isFloating()) {
    return;
  }

  Rect(getGlobalPosition(), getSize()).clip(r);
  if(Window *p = getParent()) {
    p->clip(r, this);
  }
}

void Window::fitViewport() const
{
  View *view = View::instance();
  Point gp = getGlobalPosition();

  if(!isFloating()) {
    Rect rb(gp, getSize());
    Rect r(rb);
    clip(r, this);
    Point gap = rb.getTopLeft()-r.getTopLeft();

    view->setViewport(
      Point(gp.getX()-gap.getX(), View::instance()->getSize().getHeight()-(gp.getY()+r.getHeight()-gap.getY())),
      Size(r.getWidth(), r.getHeight()));
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-gap.getX()-1, r.getWidth()+1-gap.getX(), r.getHeight()-gap.getY()+1, -gap.getY()-1, -1,5000);
    glMatrixMode(GL_MODELVIEW);
  }
  else {
    view->fitViewport();
    glTranslatef(gp.getX(), gp.getY(), 0.0f);
  }
}


void Window::update()
{
  for(window_cont::iterator p=m_children.begin(); p!=m_children.end();) {
    (*p)->update();
    ++p;
  }
}

void Window::draw() {}
float Window::getDrawPriority() const { return 1.0f; }

void Window::destroy()
{
  queueEvent(new Event(EVT_APP_DESTROY_WINDOW, this, this));
}

Window* Window::getParent() const
{
  return m_parent;
}

Window::window_cont& Window::getChildren()
{
  return m_children;
}

Window* Window::getChildByID(wnd_id id) const
{
  for(window_cont::const_iterator p=m_children.begin(); p!=m_children.end(); ++p) {
    if((*p)->getID()!=0 && (*p)->getID()==id) {
      return *p;
    }
  }
  return 0;
}

void Window::setParent(Window *parent)
{
  if(!!m_parent) {
    m_parent->deregisterChild(this);
  }
  m_parent = parent;
  if(!!parent) {
    parent->registerChild(this);
  }
}

void Window::registerChild(Window *child)
{
  if(!!child) {
    m_children.push_back(child);
  }
}

void Window::deregisterChild(Window *child)
{
  m_children.remove(child);
}

void Window::setTopLevel(Window *window)
{
  m_children.remove(window);
  m_children.push_back(window);
}

void Window::move(const Point& rel)
{
  setPosition(getPosition()+rel);
}

void Window::toTopLevel()
{
  if(getParent()) {
    getParent()->setTopLevel(this);
  }
  toTopLevelEvent();
}

void Window::toTopLevelEvent()
{
  App::instance()->getEventServer().setTopLevel(this);
  for(window_cont::iterator p=m_children.begin(); p!=m_children.end(); ++p) {
    (*p)->toTopLevelEvent();
  }
}

wnd_id Window::getID() const { return m_id; }
const Color& Window::getForegroundColor() const { return m_fg_color; }
const Color& Window::getBackgroundColor() const { return m_bg_color; }
const Color& Window::getBorderColor() const     { return m_bd_color; }
const wstring& Window::getText() const     { return m_text; }
const Point&   Window::getPosition() const      { return m_position; }
const Size&    Window::getSize() const          { return m_size; }
const Point&   Window::getClientPosition() const{ static Point dummy; return dummy; }
const Size&    Window::getClientSize() const    { return m_size; }
bool Window::isVisible() const                  { return m_visible; }
bool Window::isFloating() const                 { return m_floating; }
Point Window::getGlobalPosition() const         { return toGlobalCoord(Point()); }

void Window::setForegroundColor(const Color& color) { m_fg_color = color; }
void Window::setBackgroundColor(const Color& color) { m_bg_color = color; }
void Window::setBorderColor(const Color& color)     { m_bd_color = color; }
void Window::setText(const wstring& text)      { m_text = text; }
void Window::setPosition(const Point& pos)          { m_position = pos; }
void Window::setSize(const Size& size)              { m_size = size; }
void Window::setFloating(bool floating)             { m_floating = floating; }
void Window::setVisible(bool visible)
{
  m_visible = visible;
  for(window_cont::iterator p=m_children.begin(); p!=m_children.end(); ++p) {
    (*p)->setVisible(visible);
  }
}

Font* Window::getFont() const { return m_fontr.getFont(); }
void Window::setFont(Font *font) { m_fontr.setFont(font); }







//  Label

Label::Label(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Window(parent, pos, size, text, id)
{
}

void Label::draw()
{
//  DrawRectEdge(Rect(getSize()));
  AssignColor(getForegroundColor());
  m_fontr.draw(getText(), Point(), Rect(getPosition(), getSize()));
  AssignColor(Color());
}




//  Button

Button::Button(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Window(parent, pos, size, text, id), m_button_state(UP)
{
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_MOVE);

  if(size.getWidth()==0.0f && size.getHeight()==0.0f) {
    setSize(Size(m_fontr.getAdvance(m_text)+2.0f, m_fontr.getHeight()+2.0f));
  }
  m_fontr.setVAlign(FontRenderer::VCENTER);
  m_fontr.setHAlign(FontRenderer::HCENTER);
}

void Button::draw()
{
  if(getButtonState()==UP) {
    AssignColor(getBackgroundColor());
    DrawRect(Rect(getSize()));
  }
  else if(getButtonState()==DOWN) {
    AssignColor(Color(1.0f, 1.0f, 1.0f, 0.5f));
    DrawRect(Rect(getSize()));
  }

  AssignColor(getForegroundColor());
  m_fontr.draw(getText(), Point(), Rect(getPosition(), getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

void Button::up()
{
  setButtonState(UP);
  queueEvent(new Event(EVT_BUTTON_UP, this, getParent()));
}

void Button::down()
{
  setButtonState(DOWN);
  queueEvent(new Event(EVT_BUTTON_DOWN, this, getParent()));
}

bool Button::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(getButtonState()==UP && e.getButton()==MouseEvent::LEFT && isInside(e.getPosition())) {
      down();
      return true;
    }
  }
  else if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONUP) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(getButtonState()==DOWN && e.getButton()==MouseEvent::LEFT && isInside(e.getPosition())) {
      up();
      return true;
    }
  }
  else if(evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(getButtonState()==DOWN && !isInside(e.getPosition())) {
      setButtonState(UP);
    }
  }
  return false;
}

Button::State Button::getButtonState() const { return m_button_state; }
void Button::setButtonState(State state) { m_button_state = state; }


ToggleButton::ToggleButton(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Button(parent, pos, size, text, id), m_clicked(false)
{
}

void ToggleButton::toggle()
{
  if(getButtonState()==DOWN) {
    setButtonState(UP);
    queueEvent(new Event(EVT_BUTTON_UP, this, getParent()));
  }
  else {
    setButtonState(DOWN);
    queueEvent(new Event(EVT_BUTTON_DOWN, this, getParent()));
  }
}

bool ToggleButton::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(e.getButton()==MouseEvent::LEFT && isInside(e.getPosition())) {
      m_clicked = true;
      return true;
    }
  }
  else if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONUP) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(e.getButton()==MouseEvent::LEFT && m_clicked && isInside(e.getPosition())) {
      m_clicked = false;
      toggle();
      return true;
    }
  }
  else if(evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(m_clicked && !isInside(e.getPosition())) {
      m_clicked = false;
    }
  }
  return false;
}





//  ToolTip

ToolTip::ToolTip(Window *parent, const wstring& text, wnd_id id) :
  Window(parent, Point(), Size(), text, id), m_appear_time(0)
{
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_MOUSE_MOVE);
  listenEvent(EVT_TIMER);

  setVisible(false);
  setFloating(true);
  setSize(Size(m_fontr.getAdvance(text)+4.0f, m_fontr.getHeight(text)+4.0f));
  m_timer.setTarget(this);
}

bool ToolTip::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    setVisible(false);
  }
  else if(evt.getType()==EVT_MOUSE_BUTTONUP) {
    setVisible(false);
  }
  else if(evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);

    m_timer.stop();
    setVisible(false);
    if(getParent()->isInside(e.getPosition())) { // eWindowɃJ[\ĂA莞Ԍ\ 
      Point pos = getParent()->toLocalCoord(e.getPosition());
      pos.setY(pos.getY()+15.0f);
      setPosition(pos);
      m_timer.run(300);
      return true;
    }
  }
  else if(evt.getType()==EVT_TIMER) {
    if(evt.getDst()==this) {
      setVisible(true);
      m_appear_time = SDL_GetTicks();
      return true;
    }
  }
  return false;
}

void ToolTip::draw()
{
  float opa = std::min<float>(float(SDL_GetTicks()-m_appear_time)/100.0f, 1.0f);
  AssignColorO(getBackgroundColor(), opa);
  DrawRect(Rect(getSize()));
  AssignColorO(getForegroundColor(), opa);
  m_fontr.draw(getText(), Point(), Rect(getPosition(), getSize()));
  AssignColorO(getBorderColor(), opa);
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}


float ToolTip::getDrawPriority() const { return 2.0f; }

//  Edit

Edit::Edit(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Window(parent, pos, size, text, id), m_readonly(false), m_blinker(0), m_focused(false)
{
  listenEvent(EVT_KEYDOWN);
  listenEvent(EVT_MOUSE_BUTTONDOWN);
}

void Edit::draw()
{
  AssignColor(getBackgroundColor());
  DrawRect(Rect(getSize()));
  AssignColor(getForegroundColor());
  m_fontr.draw(getText(), Point(), Rect(getPosition(), getSize()));
  if(m_focused) {
    float x = m_fontr.getAdvance(&m_text[0], m_blinker);
    DrawLine(Point(x, 0), Point(x, getSize().getHeight()));
  }
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

void Edit::focus()
{
  m_focused = true;
}

void Edit::defocus()
{
  m_focused = false;
}

bool Edit::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(!evt.isHandled() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(e.getButton()==MouseEvent::LEFT && isInside(e.getPosition())) {
      focus();

      Point l = toLocalCoord(e.getPosition());
      float adv = 0;
      size_t i;
      for(i=0; i<m_text.size(); ++i) {
        adv+=m_fontr.getAdvance(m_text[i]);
        if(l.getX()<=adv) {
          break;
        }
      }
      setBlinkerPos(i);
      return true;
    }
    else {
      defocus();
      return false;
    }
  }
  else if(evt.getType()==EVT_KEYDOWN) {
    if(!isFocused() || isReadOnly()) {
      return false;
    }

    const KeyboardEvent& e = dynamic_cast<const KeyboardEvent&>(evt);
    wstring value = getText();
    if(e.getKey()==SDLK_RIGHT) {
      if(m_blinker<m_text.size()) {
        ++m_blinker;
      }
    }
    else if(e.getKey()==SDLK_LEFT) {
      if(m_blinker>0) {
        --m_blinker;
      }
    }
    else if(e.getKey()==SDLK_DELETE) {
      if(!m_text.empty() && m_blinker<m_text.size()) {
        value.erase(value.begin()+m_blinker);
        setText(value);
      }
    }
    else if(e.getKey()==SDLK_BACKSPACE) {
      if(!m_text.empty() && m_blinker>0 && m_blinker<=m_text.size()) {
        value.erase(value.begin()+(m_blinker-1));
        --m_blinker;
        setText(value);
      }
    }
    else if(e.getKey()==SDLK_RETURN) {
      queueEvent(new EditEvent(EVT_EDIT_ENTER, this, getParent(), getText()));
    }
    else {
      if((e.getUnicode() & 0xFF80)==0) {
        value.insert(value.begin()+m_blinker, 1, char(e.getUnicode() & 0x7F));
        ++m_blinker;
        setText(value);
      }
    }
    queueEvent(new EditEvent(EVT_EDIT, this, getParent(), getText()));
    return true;
  }
  return false;
}

void Edit::setBlinkerPos(int i)
{
  m_blinker = std::max<int>(0, std::min<int>(m_text.size(), i));
}

void Edit::setText(const wstring& text)
{
  Window::setText(text);
  if(m_blinker > text.size()) {
    m_blinker = text.size();
  }
}

bool Edit::isReadOnly() const { return m_readonly; }
bool Edit::isFocused() const { return m_focused; }
void Edit::setReadOnly(bool ro) { m_readonly = ro; }



//  DragAndDrop

DDItem::DDItem(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Window(parent, pos, size, text, id), m_dragging(false)
{
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_MOUSE_MOVE);
}

void DDItem::draw()
{
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

bool DDItem::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(!evt.isHandled() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(e.getButton()==MouseEvent::LEFT && isInside(e.getPosition())) {
      m_drag_pos = getPosition();
      m_dragging = true;
      setFloating(true);
      queueEvent(
        new DragAndDropEvent(EVT_DD_DRAG, this, getParent(),
          e.getRealPosition(), e.getRealRelative(), e.getButton(), this));
      return true;
    }
  }
  else if(evt.getType()==EVT_MOUSE_BUTTONUP) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(isDragging()) {
      m_dragging = false;
      setFloating(false);
      queueEvent(
        new DragAndDropEvent(EVT_DD_DROP, this, getParent(),
          e.getRealPosition(), e.getRealRelative(), e.getButton(), this));
      return false;
    }
  }
  else if(evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(isDragging()) {
      move(e.getRelative());
      queueEvent(new DragAndDropEvent(EVT_DD_DRAGING, this, getParent(),
        e.getRealPosition(), e.getRealRelative(), e.getButton(), this));
    }
  }
  return false;
}

bool DDItem::isDragging() const { return m_dragging; }
const Point& DDItem::getDragBeginPosition() const { return m_drag_pos; }




DDReciever::DDReciever(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id) :
  Window(parent, pos, size, text, id), m_item(0)
{
  listenEvent(EVT_DD_DROP);
}

void DDReciever::draw()
{
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

bool DDReciever::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(evt.getType()==EVT_DD_DROP) {
    const DragAndDropEvent& e = dynamic_cast<const DragAndDropEvent&>(evt);
    if(!getItem() && isInside(e.getPosition())) {
      if(DDItem *item = dynamic_cast<DDItem*>(e.getSrc())) {
        setItem(item);
        queueEvent(new DragAndDropEvent(EVT_DD_RECIEVE, this, getParent(),
          e.getRealPosition(), e.getRealRelative(), e.getButton(), item));
        return true;
      }
    }
  }
  return false;
}

void DDReciever::deregisterChild(Window *child)
{
  Window::deregisterChild(child);
  if(child==m_item) {
    setItem(0);
  }
}

void DDReciever::setItem(DDItem *item)
{
  m_item = item;
  if(m_item) {
    Size gap = getSize()-m_item->getSize();
    m_item->setPosition(Point(gap.getWidth()/2.0f, gap.getHeight()/2.0f));
    m_item->setParent(this);
  }
}
DDItem* DDReciever::getItem() const { return m_item; }




//  ScrollBar

ScrollBar::ScrollBar(Window *parent, int style, float size, wnd_id id) :
  Window(parent, Point(), Size(), L"", id),
  m_align(HORIZONTAL), m_dragging(false), m_thick(0.0f), m_step(0.0f), m_scroll_pos(0.0f), m_range(0.0f), m_bar_width(10.0f),
  m_button_tl(0), m_button_br(0)
{
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_MOUSE_MOVE);
  listenEvent(EVT_BUTTON_UP);

  setThick(12.0f);
  setStep(10.0f);
  setRange(100.0f);
  setValue(0.0f);
  resetBarRect();

  if(style & SB_VERTICAL) {
    m_align = VERTICAL;
  }

  if(m_align==HORIZONTAL) {
    m_button_tl = new Button(this, Point(), Size(getThick(), getThick()));
    m_button_br = new Button(this, Point(getSize().getWidth()-getThick(), 0.0f), Size(getThick(), getThick()));
  }
  else if(m_align==VERTICAL) {
    m_button_tl = new Button(this, Point(), Size(getThick(), getThick()));
    m_button_br = new Button(this, Point(0.0f, getSize().getHeight()-getThick()), Size(getThick(), getThick()));
  }

}

void ScrollBar::draw()
{
  AssignColor(Color(1.0f, 1.0f, 1.0f, 0.5f));
  DrawRect(m_bar);
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

bool ScrollBar::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(evt.getType()==EVT_BUTTON_UP) {
    if(!!m_button_tl && evt.getSrc()==m_button_tl) {
      setValue(getValue()-getStep());
      return true;
    }
    else if(m_button_br && evt.getSrc()==m_button_br) {
      setValue(getValue()+getStep());
      return true;
    }
  }
  else if(!evt.isHandled() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(m_bar.isInside(toLocalCoord(e.getPosition()))) {
      m_dragging = true;
      return true;
    }
  }
  else if(evt.getType()==EVT_MOUSE_BUTTONUP) {
    m_dragging = false;
  }
  else if(evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(m_dragging) {
      mouseToScroll(e.getRelative());
    }
  }
  return Window::handleEvent(evt);
}

void ScrollBar::fitTo(Window *target)
{
  if(m_align==HORIZONTAL) {
    setPosition(Point(0.0f, target->getSize().getHeight()-getThick()));
    setSize(Size(target->getSize().getWidth()-getThick(), getThick()));
    m_button_tl->setPosition(Point());
    m_button_br->setPosition(Point(getSize().getWidth()-getThick(), 0.0f));
  }
  else if(m_align==VERTICAL) {
    setPosition(Point(target->getSize().getWidth()-getThick(), 0.0f));
    setSize(Size(getThick(), target->getSize().getHeight()-getThick()));
    m_button_tl->setPosition(Point());
    m_button_br->setPosition(Point(0.0f, getSize().getHeight()-getThick()));
  }
}

float ScrollBar::getBarWidth() const
{
  return m_bar_width;
}

Range ScrollBar::getScrollBarRange() const
{
  Range r;
  if(m_align==HORIZONTAL) {
    r.setMin(getThick());
    r.setMax(getSize().getWidth()-getThick()-m_bar_width);
  }
  else if(m_align==VERTICAL) {
    r.setMin(getThick());
    r.setMax(getSize().getHeight()-getThick()-m_bar_width);
  }
  return r;
}

void ScrollBar::resetBarRect()
{
  Range r = getScrollBarRange();
  if(m_align==HORIZONTAL) {
    float width = getSize().getWidth()-getThick()*2.0f;
    float par = std::min<float>(1.0f, getParent()->getSize().getWidth()/getRange());
    m_bar_width = std::max<float>(10.0f, width*par);
    m_bar = Rect(Point(r.getMin()+r.getLength()*(getValue()/getRange()), 0.0f), Size(m_bar_width, getThick()));
  }
  else if(m_align==VERTICAL) {
    float width = getSize().getHeight()-getThick()*2.0f;
    float par = std::min<float>(1.0f, getParent()->getSize().getHeight()/getRange());
    m_bar_width = std::max<float>(10.0f, width*par);
    m_bar = Rect(Point(0.0f, r.getMin()+r.getLength()*(getValue()/getRange())), Size(getThick(), m_bar_width));
  }
}

void ScrollBar::mouseToScroll(const Point& mouse_move)
{
  Range r = getScrollBarRange();
  if(m_align==HORIZONTAL) {
    setValue(getValue()+mouse_move.getX()/r.getLength()*getRange());
  }
  else if(m_align==VERTICAL) {
    setValue(getValue()+mouse_move.getY()/r.getLength()*getRange());
  }
}

float ScrollBar::getValue() const { return m_scroll_pos; }
float ScrollBar::getThick() const { return m_thick; }
float ScrollBar::getStep() const  { return m_step; }
float ScrollBar::getRange() const { return m_range; }

void ScrollBar::setScrollBarSize(float size)
{
  if(m_align==HORIZONTAL) {
    setSize(Size(size, getThick()));
    m_button_tl->setPosition(Point());
    m_button_br->setPosition(Point(getSize().getWidth()-getThick(), 0.0f));
  }
  else if(m_align==VERTICAL) {
    setSize(Size(getThick(), size));
    m_button_tl->setPosition(Point());
    m_button_br->setPosition(Point(0.0f, getSize().getHeight()-getThick()));
  }
  resetBarRect();
}

void ScrollBar::setValue(float value)
{
  float old_value = m_scroll_pos;
  m_scroll_pos = std::max<float>(0.0f, std::min<float>(value, getRange()));
  resetBarRect();

  float scale = 1.0f;
  if(m_align==HORIZONTAL) {
    scale = std::max<float>(0.0f, (getRange()-getParent()->getSize().getWidth())/getRange());
  }
  else if(m_align==VERTICAL) {
    scale = std::max<float>(0.0f, (getRange()-getParent()->getSize().getHeight())/getRange());
  }
  queueEvent(new ScrollEvent(EVT_SCROLL, this, getParent(), m_scroll_pos*scale, (m_scroll_pos-old_value)*scale));
    
}

void ScrollBar::setRange(float range)
{
  float old = getValue();
  setValue(0.0f);
  m_range = range;
  setValue(old);
}
void ScrollBar::setThick(float thick) { m_thick = thick; }
void ScrollBar::setStep(float value) { m_step = value; }



//  ScrolledWindow

ScrolledWindow::ScrolledWindow(Window *parent, const Point& pos, const Size& size,
  const wstring& text, wnd_id id, int style) :
  Window(parent, pos, size, text, id), m_hscrollbar(0), m_vscrollbar(0)
{
  if(style & VSCROLLBAR) {
    m_vscrollbar = new ScrollBar(this, SB_VERTICAL);
    m_vscrollbar->setPosition(Point(getSize().getWidth()-m_vscrollbar->getThick(), 0.0f));
    if(style & HSCROLLBAR)
      m_vscrollbar->setScrollBarSize(getSize().getHeight()-m_vscrollbar->getThick());
    else
      m_vscrollbar->setScrollBarSize(getSize().getHeight());
  }
  if(style & HSCROLLBAR) {
    m_hscrollbar = new ScrollBar(this, SB_HORIZONTAL);
    m_hscrollbar->setPosition(Point(0.0f, getSize().getHeight()-m_hscrollbar->getThick()));
    if(style & VSCROLLBAR)
      m_hscrollbar->setScrollBarSize(getSize().getWidth()-m_hscrollbar->getThick());
    else
      m_hscrollbar->setScrollBarSize(getSize().getWidth());
  }
  setSize(getSize());

  listenEvent(EVT_SCROLL);
}

bool ScrolledWindow::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(evt.getType()==EVT_SCROLL) {
    if((!!m_vscrollbar && evt.getSrc()==m_vscrollbar) || (!!m_hscrollbar && evt.getSrc()==m_hscrollbar)) {
      const ScrollEvent& se = dynamic_cast<const ScrollEvent&>(evt);
      if(!!m_vscrollbar && se.getSrc()==m_vscrollbar) {
        scroll(Point(0.0f, -se.getRelative()));
        return true;
      }
      else if(!!m_hscrollbar && se.getSrc()==m_hscrollbar) {
        scroll(Point(-se.getRelative(), 0.0f));
        return true;
      }
    }
  }
  return false;
}

void ScrolledWindow::draw()
{
  AssignColor(getBackgroundColor());
  DrawRect(Rect(Point(), Window::getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(Point(), Window::getSize()));
  AssignColor(Color());
}

void ScrolledWindow::scroll(const Point& rel)
{
  m_base_pos+=rel;
  for(window_cont::iterator p=m_children.begin(); p!=m_children.end(); ++p) {
    if(*p!=m_vscrollbar && *p!=m_hscrollbar)
      (*p)->move(rel);
  }
}

const Point& ScrolledWindow::getBasePosition() const { return m_base_pos; }
const Size& ScrolledWindow::getSizeNS() const { return m_size_ns; }
const Size& ScrolledWindow::getScrollSize() const { return m_scroll_size; }
ScrollBar* ScrolledWindow::getVScrollBar() const { return m_vscrollbar; }
ScrollBar* ScrolledWindow::getHScrollBar() const { return m_hscrollbar; }

void ScrolledWindow::setSize(const Size& size)
{
  Window::setSize(size);
  m_size_ns = getSize();
  if(!!m_hscrollbar) {
    m_size_ns.setHeight(m_size_ns.getHeight()-12.0f);
  }
  if(!!m_vscrollbar) {
    m_size_ns.setWidth(m_size_ns.getWidth()-12.0f);
  }
}

void ScrolledWindow::setScrollSize(const Size& size)
{
  m_scroll_size = size;
  if(!!m_hscrollbar) {
    m_hscrollbar->setRange(getScrollSize().getWidth());
  }
  if(!!m_vscrollbar) {
    m_vscrollbar->setRange(getScrollSize().getHeight());
  }
}

void ScrolledWindow::clip(Rect& r, const Window *wnd) const
{
  if(!!wnd && wnd->isFloating()) {
    return;
  }

  if(!!wnd && (wnd==this || wnd==m_vscrollbar || wnd==m_hscrollbar)) {
    Window::clip(r, wnd);
  }
  else {
    Rect(getGlobalPosition(), getSizeNS()).clip(r);
    if(Window *p = getParent())
      p->clip(r, this);
  }
}




//  List

ListItem::ListItem(const wstring& text) :
  Window(0, Point(), Size(), text, 0), m_selected(false)
{
}

void ListItem::draw()
{
  if(isSelected()) {
    AssignColor(Color(1.0f, 1.0f, 1.0f, 0.5f));
    DrawRect(Rect(getSize()));
  }
  AssignColor(getForegroundColor());
  m_fontr.draw(getText(), Point(), Rect(getPosition(), getSize()));
  AssignColor(Color());
}

bool ListItem::isSelected() const { return m_selected; }
void ListItem::setSelect(bool selected) { m_selected = selected; }




List::List(Window *parent, const Point& pos, const Size& size, wnd_id id, int style) :
  ScrolledWindow(parent, pos, size, L"", id, VSCROLLBAR),
  m_multiselect(false), m_item_height(12.0f), m_last_selected(0)
{
  m_item_height = std::ceil(App::instance()->getDefaultFont()->LineHeight());
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_TIMER);
}

void List::draw()
{
  AssignColor(getBackgroundColor());
  DrawRect(Rect(getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}


bool List::handleEvent(const Event& evt)
{
  if(!isVisible()) {
    return false;
  }

  if(!evt.isHandled() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(e.getButton()==MouseEvent::LEFT && getItemCount()!=0 && isInside(e.getPosition())) {
      size_t index = hitItem(e.getPosition());
      if(ListItem *item = getItem(index)) {
        if(m_stopwatch.isRunning() && item==m_last_selected) {
          queueEvent(new ListEvent(EVT_LIST_DOUBLECLICK, this, getParent(), index, item));
          m_stopwatch.stop();
          m_last_selected = 0;
        }
        else {
          selectItem(index);
          m_last_selected = item;
          m_stopwatch.run(300); // _uNbNoJn 
        }
      }
      return true;
    }
  }
  return ScrolledWindow::handleEvent(evt);
}

void List::selectItem(size_t i)
{
  if(i>=m_items.size()) {
    return;
  }

  if(!m_multiselect) {
    clearSelection();
  }

  ListItem *item = m_items[i];
  if(!item->isSelected()) {
    item->setSelect(true);
    queueEvent(new ListEvent(EVT_LIST_SELECT, this, getParent(), i, item));
  }
  else if(m_multiselect) {
    item->setSelect(false);
    queueEvent(new ListEvent(EVT_LIST_DISSELECT, this, getParent(), i, item));
  }
}

void List::selectItem(ListItem *l)
{
  selectItem(getItemIndex(l));
}

size_t List::hitItem(const Point& p) const
{
  size_t i;
  for(i=0; i<m_items.size(); ++i) {
    if(m_items[i]->isInside(p)) {
      break;
    }
  }
  return i;
}

ListItem* List::getItem(size_t i)
{
  if(i>=m_items.size()) {
    return 0;
  }
  return m_items[i];
}

ListItem* List::getSelectedItem(size_t i)
{
  if(i < getSelectionCount()) {
    return getItem(getSelection(i));
  }
  return 0;
}

size_t List::getItemCount() const
{
  return m_items.size();
}

size_t List::getItemIndex(ListItem *l) const
{
  size_t i = 0;
  for(i=0; i<m_items.size(); ++i) {
    if(l==m_items[i]) {
      break;
    }
  }
  return i;
}

size_t List::getSelectionCount() const
{
  size_t count = 0;
  for(size_t i=0; i<m_items.size(); ++i) {
    if(m_items[i]->isSelected()) {
      ++count;
    }
  }
  return count;
}

size_t List::getSelection(size_t index) const
{
  size_t count = 0;
  size_t i;
  for(i=0; i<m_items.size(); ++i) {
    if(m_items[i]->isSelected()) {
      if(count==index) {
        return i;
      }
      ++count;
    }
  }
  return i;
}

void List::addItem(ListItem *item)
{
  if(!!item) {
    item->setParent(this);
    m_items.push_back(item);
    alignItems();
    resetScrollSize();
  }
}

void List::removeItem(size_t i)
{
  if(i>=m_items.size()) {
    return;
  }
  m_items[i]->destroy();
  m_items.erase(m_items.begin()+i);
  resetScrollSize();
}

void List::removeItem(ListItem *l)
{
  removeItem(getItemIndex(l));
}

void List::clearItem()
{
  for(item_cont::iterator p=m_items.begin(); p!=m_items.end(); ++p) {
    (*p)->destroy();
  }
  m_items.clear();
  resetScrollSize();
}

void List::clearSelection()
{
  for(size_t i=0; i<m_items.size(); ++i) {
    if(m_items[i]->isSelected()) {
      m_items[i]->setSelect(false);
      queueEvent(new ListEvent(EVT_LIST_DISSELECT, this, getParent(), i, m_items[i]));
    }
  }
}

void List::alignItems()
{
  for(size_t i=0; i<m_items.size(); ++i) {
    m_items[i]->setPosition(getBasePosition()+Point(0.0f, getItemHeight()*i));
    m_items[i]->setSize(Size(getSize().getWidth(), getItemHeight()));
  }
}

void List::resetScrollSize()
{
  setScrollSize(Size(getSize().getWidth(), getItemHeight()*getItemCount()+2.0f));
}

float List::getItemHeight() const { return m_item_height; }




//  Combo

Combo::Combo(Window *parent, const Point& pos, const Size& size, wnd_id id, int style) :
  Window(parent, pos, size, L"", id), m_list_button(0), m_list(0)
{
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_BUTTON_DOWN);
  listenEvent(EVT_LIST_SELECT);

  m_list_button = new Button(this, Point(getSize().getWidth()-16.0f, 0.0f), Size(16.0f, 16.0f));
  m_list = new List(this, Point(0.0f, getSize().getHeight()), Size(getSize().getWidth(), 150.0f));
  m_list->setFloating(true);
  m_list->setVisible(false);
}

bool Combo::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_BUTTON_DOWN) {
    if(!!m_list_button && evt.getSrc()==m_list_button) {
      if(m_list->isVisible()) {
        m_list->setVisible(false);
      }
      else {
        m_list->setVisible(true);
      }
      return true;
    }
  }
  else if(evt.getType()==EVT_LIST_SELECT) {
    if(!!m_list && evt.getSrc()==m_list) {
      setSelection(m_list->getSelection());
      m_list->setVisible(false);
      return true;
    }
  }

  if(!isVisible()) {
    return false;
  }

  if(evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(!m_list_button->isInside(e.getPosition())) {
      if(m_list->isVisible() && !m_list->isInside(e.getPosition())) {
        m_list->setVisible(false);
        return true;
      }
      else if(!evt.isHandled() && !m_list->isVisible() && isInside(e.getPosition())) {
        m_list->setVisible(true);
        return true;
      }
    }
  }
  else if(evt.getType()==EVT_MOUSE_BUTTONUP) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(!m_list_button->isInside(e.getPosition())) {
      if(m_list->isVisible() && !isInside(e.getPosition()) && !m_list->isInside(e.getPosition())) {
        m_list->setVisible(false);
        return true;
      }
    }
  }
  return Window::handleEvent(evt);
}

void Combo::draw()
{
  AssignColor(getForegroundColor());
  m_fontr.draw(getText(), Point(), Rect(getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  AssignColor(Color());
}

void Combo::addItem(ListItem *item)
{
  m_list->addItem(item);
}

void Combo::clearItem()
{
  m_list->clearItem();
  m_selection = 0;
}

List* Combo::getList() { return m_list; }
ListItem* Combo::getSelection() { return m_selection; }

void Combo::setSelection(size_t i)
{
  m_selection = m_list->getItem(i);
  assert(!!m_selection);
  setText(m_selection->getText());
  queueEvent(new ListEvent(EVT_COMBO, this, getParent(), i, m_selection));
}





//  Panel

Panel::Panel(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id, int style) :
  Window(parent, pos, size, text, id)
{
}

void Panel::draw()
{
  AssignColor(getBackgroundColor());
  DrawRect(Rect(getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));

  AssignColor(Color());
}

bool Panel::handleEvent(const Event& evt)
{
  return false;
}

void Panel::setSize(const Size& size)
{
  Size old = getSize();
  Window::setSize(size);
  queueEvent(new ResizeEvent(EVT_RESIZE, this, getParent(), getSize(), getSize()-old));
}




//  Dialog

Dialog::Dialog(Window *parent, const Point& pos, const Size& size, const wstring& text, wnd_id id, int style) :
  Panel(parent, pos, size, text, id, style),
  m_close_button(0), m_maximize_button(0), m_minimize_button(0),
  m_movable(false), m_moving(false),
  m_resizable(false), m_resizing(0)
{
  listenEvent(EVT_MOUSE_BUTTONDOWN);
  listenEvent(EVT_MOUSE_BUTTONUP);
  listenEvent(EVT_MOUSE_MOVE);
  listenEvent(EVT_BUTTON_UP);

  if(getParent()==0) {
    setParent(View::instance());
  }
  setWidthRange(Range(120.0f, getParent()->getSize().getWidth()));
  setHeightRange(Range(25.0f, getParent()->getSize().getHeight()));
//  m_client_pos = Point(0.0f, 24.0f);

  if(style & MOVABLE) {
    m_movable = true;
  }
  if(style & RESIZABLE) {
    m_resizable = true;
  }

  int buttons = 1;
  if(style & CLOSE_BUTTON) {
    m_close_button = new Button(this, Point(size.getWidth()-20*buttons, 4), Size(16, 16));
    ++buttons;
  }
  if(style & MAXIMIZE_BUTTON) {
    m_maximize_button = new Button(this, Point(size.getWidth()-20*buttons, 4), Size(16, 16));
    ++buttons;
  }
  if(style & MINIMIZE_BUTTON) {
    m_minimize_button = new Button(this, Point(size.getWidth()-20*buttons, 4), Size(16, 16));
    ++buttons;
  }
}

void Dialog::destroy()
{
  queueEvent(new Event(EVT_CLOSEDIALOG, this, getParent()));
  Panel::destroy();
}

void Dialog::draw()
{
  AssignColor(getBackgroundColor());
  DrawRect(Rect(getSize()));
  AssignColor(getBorderColor());
  DrawRectEdge(Rect(getSize()));
  DrawLine(Point(0, 24), Point(getSize().getWidth(), 24));
  m_fontr.draw(getText(), Point(4,4), Rect(getPosition(), Size(getSize().getWidth(), 24)));

  AssignColor(Color());
}

void Dialog::onCloseButton()
{
  destroy();
}

void Dialog::onMaximizeButton()
{
  setSize(Size(m_width_range.getMax(), m_height_range.getMax()));
  setPosition(Point(0.0f, 0.0f));
}

void Dialog::onMinimizeButton()
{
  setSize(Size(m_width_range.getMin(), m_height_range.getMin()));
}

bool Dialog::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_BUTTON_UP) {
    if(!!m_close_button && evt.getSrc()==m_close_button) {
      onCloseButton();
      return true;
    }
    else if(!!m_maximize_button && evt.getSrc()==m_maximize_button) {
      onMaximizeButton();
      return true;
    }
    else if(!!m_minimize_button && evt.getSrc()==m_minimize_button) {
      onMinimizeButton();
      return true;
    }
  }

  if(!isVisible()) {
    return Panel::handleEvent(evt);
  }

  if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONDOWN) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(isInside(e.getPosition())) {
      toTopLevel();

      if(e.getButton()==MouseEvent::LEFT) {
        const Size& s = getSize();
        Point gp = getGlobalPosition();

        // ^Cgo[NbNꂽ 
        if(Rect(gp, Size(getSize().getWidth(), 24)).isInside(e.getPosition())) {
          if(m_movable && !evt.isHandled()) {
            m_moving = true;
          }
        }
        else {
          // [NbNꂽ 
          if(Rect(gp, Size(5, s.getHeight())).isInside(e.getPosition())) {
            m_resizing|=1;
          }
          else if(Rect(gp+Point(s.getWidth()-5.0f, 0.0f), Size(5.0f, s.getHeight())).isInside(e.getPosition())) {
            m_resizing|=2;
          }

          if(Rect(gp+Point(0.0f, s.getHeight()-5.0f), Size(s.getWidth(), 5.0f)).isInside(e.getPosition())) {
            m_resizing|=4;
          }
        }
      }
      return true;
    }
  }
  else if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_BUTTONUP) {
    m_moving = false;
    m_resizing = 0;

    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(isInside(e.getPosition())) {
      return true;
    }
  }
  else if(!evt.isHandledByFrame() && evt.getType()==EVT_MOUSE_MOVE) {
    const MouseEvent& e = dynamic_cast<const MouseEvent&>(evt);
    if(m_moving) { // ^Cgo[hbOňʒuύX 
      setPosition(getPosition()+e.getRelative());
    }
    if(m_resizing!=0) { // [hbOŃEBhETCYύX 
      Size size = getSize();
      Point pos = getPosition();
      if(m_resizing&1) { //  
        size.setWidth(size.getWidth()-e.getRelative().getX());
        pos.setY(pos.getY()+e.getRelative().getX());
      }
      else if(m_resizing&2) { // E 
        size.setWidth(size.getWidth()+e.getRelative().getX());
      }
      if(m_resizing&4) { //  
        size.setHeight(size.getHeight()+e.getRelative().getY());
      }
      size.setWidth(m_width_range.clamp(size.getWidth()));
      size.setHeight(m_height_range.clamp(size.getHeight()));
      setSize(size);
      setPosition(pos);

      if(isInside(e.getPosition())) {
        return true;
      }
    }
  }
  return Panel::handleEvent(evt);
}

const Point& Dialog::getClientPosition() const { return m_client_pos; }
const Size& Dialog::getClientSize() const { return m_client_size; }

void Dialog::setSize(const Size& size)
{
  Size old = getSize();
  Panel::setSize(size);

  Size gap = getSize()-old;
  Point rel(gap.getWidth(), 0.0f);
  if(!!m_close_button) { m_close_button->move(rel); }
  if(!!m_minimize_button) { m_minimize_button->move(rel); }
  if(!!m_maximize_button) { m_maximize_button->move(rel); }
}

void Dialog::setWidthRange(const Range& r) { m_width_range = r; }
void Dialog::setHeightRange(const Range& r) { m_height_range = r; }




// MessageDialog

MessageDialog::MessageDialog(const wstring& title, const wstring& message, Window *parent, const Point& pos, const Size& size):
  Dialog(parent, pos, size, title, 0, MOVABLE | CLOSE_BUTTON),
  m_ok_button(0), m_message(0)
{
  m_ok_button = new Button(this, Point(size.getWidth()-55, size.getHeight()-25), Size(50, 20), L"OK");
  m_message = new Label(this, Point(5,30), Size(size-Size(10, 35)), message);
}

bool MessageDialog::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_BUTTON_UP) {
    if(evt.getDst()==this && evt.getSrc()==m_ok_button) {
      destroy();
      return true;
    }
  }
  return Dialog::handleEvent(evt);
}


// ConfirmDialog

ConfirmDialog::ConfirmDialog(const wstring& title, const wstring& message, Window *parent, const Point& pos, const Size& size):
  Dialog(parent, pos, size, title, 0, MOVABLE | CLOSE_BUTTON),
  m_ok_button(0), m_cancel_button(0), m_message(0)
{
  m_ok_button = new Button(this, Point(), Size(), L"OK");
  m_cancel_button = new Button(this, Point(), Size(), L"Cancel");
  m_message = new Label(this, Point(), Size(), message);
}

bool ConfirmDialog::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_BUTTON_UP) {
    if(evt.getSrc()==m_ok_button || evt.getSrc()==m_cancel_button) {
      destroy();
      if(evt.getSrc()==m_ok_button) {
        queueEvent(new DialogEvent(EVT_CONFIRMDIALOG, this, getParent(), true));
      }
      else {
        queueEvent(new DialogEvent(EVT_CONFIRMDIALOG, this, getParent(), false));
      }
      return true;
    }
  }
  return Dialog::handleEvent(evt);
}






// FileDialog

#ifdef SG_WITH_FILEDIALOG




class CWConverter
{
public:
  static CWConverter& getDefault()
  {
    static CWConverter c;
    return c;
  }

  CWConverter(const char *l = "") {
    setLocale(l);
  }

  void setLocale(const char *l)
  {
    loc = l;
    ::setlocale(LC_ALL, l);
  }

  void MBS2WCS(wstring& dst, const string& src)
  {
    size_t len = mbstowcs(0, src.c_str(), 0)+1;
    if(len==(size_t)(-1)) {
      return;
    }
    wchar_t *buf(new wchar_t[len]);
    ::mbstowcs(buf, src.c_str(), len);
    dst = buf;
    delete[] buf;
  }

  void WCS2MBS(string& dst, const wstring& src)
  {
    size_t len = wcstombs(0, src.c_str(), 0)+1;
    if(len==(size_t)(-1)) {
      return;
    }
    char *buf(new char[len]);
    ::wcstombs(buf, src.c_str(), len);
    dst = buf;
    delete[] buf;
  }

  string loc;
};

wstring _L(const string& src)
{
  wstring dst;
  CWConverter::getDefault().MBS2WCS(dst, src);
  return dst;
}

string _S(const wstring& src)
{
  string dst;
  CWConverter::getDefault().WCS2MBS(dst, src);
  return dst;
}



bool IsFile(const string& path)
{
#ifdef _WIN32
  DWORD ret = ::GetFileAttributes(path.c_str());
  return (ret!=(DWORD)-1) && !(ret & FILE_ATTRIBUTE_DIRECTORY);
#else
  struct stat st;
  return ::stat(path.c_str(), &st)==0 && S_ISREG(st.st_mode);
#endif
}

bool IsDir(const string& path)
{
#ifdef _WIN32
  DWORD ret = ::GetFileAttributes(path.c_str());
  return (ret!=(DWORD)-1) && (::GetFileAttributes(path.c_str())&FILE_ATTRIBUTE_DIRECTORY)!=0;
#else
  struct stat st;
  return ::stat(path.c_str(), &st)==0 && S_ISDIR(st.st_mode);
#endif
}

bool MakeDir(const string& path)
{
#ifdef _WIN32
  return ::CreateDirectory(path.c_str(), NULL)==TRUE;
#else
  return ::mkdir(path.c_str(), 0777)==0;
#endif
}

bool MakeDeepDir(const string& path)
{
  size_t num = 0;
  for(size_t i=0; i<path.size(); ++i) {
    if(path[i]=='/') {
      string p(path.begin(), path.begin()+i);
      if(MakeDir(p.c_str())) {
        ++num;
      }
    }
  }
  return num!=0;
}

bool RemoveDir(const string& path)
{
#ifdef _WIN32
  return ::RemoveDirectory(path.c_str())==TRUE;
#else
  return ::rmdir(path.c_str())==0;
#endif
}

bool Remove(const string& path)
{
  if(IsDir(path)) {
    return RemoveDir(path.c_str());
  }
  else {
    return std::remove(path.c_str())==0;
  }
}

bool RemoveRecursive(const string& path)
{
  bool res = false;
  if(IsDir(path)) {
    Dir dir(path);
    for(Dir::iterator p=dir.begin(); p!=dir.end(); ++p) {
      if(*p=="." || *p=="..") {
        continue;
      }
      string next = dir.getPath()+*p;
      res = RemoveRecursive(next.c_str());
    }
    res = RemoveDir(path);
  }
  else {
    res = std::remove(path.c_str())==0;
  }
  return res;
}

string GetCWD()
{
  char buf[256];
#ifdef _WIN32
  ::GetCurrentDirectory(256, buf);
  for(size_t i=0; i<strlen(buf); ++i) {
    if(buf[i]=='\\') {
      buf[i] = '/';
    }
  }
  return buf;
#else
  ::getcwd(buf, 256);
  return buf;
#endif
}

bool SetCWD(const string& path)
{
#ifdef _WIN32
  return ::SetCurrentDirectory(path.c_str())!=0;
#else
  return ::chdir(path.c_str())==0;
#endif
}


Dir::Dir() {}
Dir::Dir(const string& path) { open(path); }
size_t Dir::size() const { return m_files.size(); }
const string& Dir::operator[](size_t i) const { return m_files[i]; }
const string& Dir::getPath() const { return m_path; }

bool Dir::open(const string& path)
{
  m_files.clear();
  m_path = path;
  if(m_path[m_path.size()-1]!='/') {
    m_path+='/';
  }

#ifdef _WIN32
  WIN32_FIND_DATA wfdata;
  HANDLE handle = ::FindFirstFile((m_path+"*").c_str(), &wfdata);
  if(handle!=INVALID_HANDLE_VALUE) {
    do {
      m_files.push_back(wfdata.cFileName);
    } while(::FindNextFile(handle, &wfdata));
    ::FindClose(handle);
  }
  else {
    return false;
  }
  return true;
#else
  DIR *dir = ::opendir(m_path.c_str());
  if(dir!=0) {
    dirent *dr;
    while((dr=::readdir(dir))!=0) {
      m_files.push_back(dr->d_name);
    }
    ::closedir(dir);
  }
  else {
    return false;
  }
  return true;
#endif
}

bool Dir::openRecursive(const string& path)
{
  Dir dir(path);
  for(iterator p=dir.begin(); p!=dir.end(); ++p) {
    if(*p=="." || *p=="..") {
      continue;
    }
    string next = dir.getPath()+*p;
    if(IsDir(next.c_str())) {
      openRecursive(next.c_str());
    }
    else {
      m_files.push_back(next);
    }
  }
  return size()!=0;
}



FileDialog::FileDialog(const wstring& title, Window *parent, const Point& pos, const Size& size):
  Dialog(parent, pos, size, title, 0, MOVABLE | CLOSE_BUTTON),
  m_ok_button(0), m_cancel_button(0), m_filelist(0)
{
  setSize(Size(320, 250));
  m_ok_button = new Button(this, Point(260, 30), Size(50, 16), L"OK");
  m_cancel_button = new Button(this, Point(260, 50), Size(50, 16), L"Cancel");
  m_filelist = new List(this, Point(5, 50), Size(250, 180));
  m_locate = new Combo(this, Point(5, 30), Size(250, 16));
  openDir(GetCWD());
}

bool FileDialog::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_BUTTON_UP) {
    if(evt.getSrc()==m_ok_button || evt.getSrc()==m_cancel_button) {
      if(evt.getSrc()==m_ok_button) {
        queueEvent(new DialogEvent(EVT_FILEDIALOG, this, getParent(), true));
      }
      else {
        queueEvent(new DialogEvent(EVT_FILEDIALOG, this, getParent(), false));
      }
      destroy();
      return true;
    }
  }
  else if(evt.getType()==EVT_LIST_DOUBLECLICK) {
    if(evt.getDst()==this) {
      const ListEvent& e = dynamic_cast<const ListEvent&>(evt);
      if(IsDir(m_path+m_file)) {
        openDir(m_path+m_file);
      }
      else if(IsFile(m_path+m_file)) {
        queueEvent(new DialogEvent(EVT_FILEDIALOG, this, getParent(), true, m_path+m_file));
        destroy();
      }
      return true;
    }
  }
  else if(evt.getType()==EVT_LIST_SELECT) {
    if(evt.getDst()==this) {
      const ListEvent& e = dynamic_cast<const ListEvent&>(evt);
      m_file = _S(e.getItem()->getText());
      return true;
    }
  }
  else if(evt.getType()==EVT_COMBO) {
    if(evt.getDst()==this) {
      const ListEvent& e = dynamic_cast<const ListEvent&>(evt);
      openDir(_S(e.getItem()->getText()));
      return true;
    }
  }
  return Dialog::handleEvent(evt);
}

bool FileDialog::openDir(const string& _path)
{
  string path = _path;

  if(path.empty()) {
    return false;
  }
  if(path[path.size()-1]!='/') {
    path+='/';
  }
  if(!IsDir(path) || m_path==path) {
    return false;
  }

  if(path.size()>4 && strncmp(path.c_str()+path.size()-3, "/./", 3)==0) {
    return true;
  }
  else if(path.size()>5 && strncmp(path.c_str()+path.size()-4, "/../", 4)==0) {
    string p = path;
    size_t s = p.size()-5;
    for(; s!=0; --s) {
      if(p[s]=='/') {
        break;
      }
    }
    p.erase(p.begin()+s, p.end());
    m_path = p;
  }
  else {
    m_path = path;
  }

  if(m_path[m_path.size()-1]!='/') {
    m_path+='/';
  }

  m_locate->clearItem();
  for(int i=int(m_path.size()-1); i>=0; --i) {
    if(m_path[i]=='/' || m_path[i]=='\\') {
      m_locate->addItem(new ListItem(_L(string(m_path.begin(), m_path.begin()+i))));
    }
  }
  m_locate->setSelection(0);

  m_file.clear();
  m_filelist->clearItem();
  Dir dir(path);
  for(size_t i=0; i<dir.size(); ++i) {
    if(IsDir(m_path+dir[i])) {
      m_filelist->addItem(new ListItem(_L(dir[i]+'/')));
    }
  }
  for(size_t i=0; i<dir.size(); ++i) {
    if(!IsDir(m_path+dir[i])) {
      m_filelist->addItem(new ListItem(_L(dir[i])));
    }
  }
  return true;
}

#endif







/*
  View
*/

View* View::m_instance = 0;

View* View::instance()
{
  return m_instance;
}

View::View(const string& title, const Size& window_size, const Size& gl_size, bool full) :
  Window(0), m_width_scale(1.0f), m_height_scale(1.0f), m_fullscreen(full)
{
  assert(m_instance==0);

  m_instance = this;
  if(window_size.getWidth()>0.0f) {
    m_width_scale = gl_size.getWidth()/window_size.getWidth();
    m_height_scale = gl_size.getHeight()/window_size.getHeight();
  }

  setSize(window_size);
  SDL_WM_SetCaption(title.c_str(), 0);
}

void View::setSize(const Size& size)
{
  Window::setSize(Size(size.getWidth()*m_width_scale, size.getHeight()*m_height_scale));
  m_window_size = size;

  int flag = SDL_OPENGL;
  int bpp = 0;
  if(m_fullscreen) {
    flag|=SDL_FULLSCREEN;
    bpp = 32;
  }

  if(SDL_SetVideoMode(int(size.getWidth()), int(size.getHeight()), bpp, flag)==0) {
    throw std::runtime_error(SDL_GetError());
  }


  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glShadeModel(GL_SMOOTH);
  glCullFace(GL_BACK);
  glFrontFace(GL_CCW);
  glEnable(GL_CULL_FACE);
  glEnable(GL_BLEND);
  glClearColor(0, 0, 0, 0);

  App::instance()->initOpenGLResource();
}

namespace {
  typedef std::vector<Window*> draw_queue;
  draw_queue g_draw;

  void gAdd(Window *w)
  {
    if(!w->isVisible()) {
      return;
    }

    g_draw.push_back(w);
    Window::window_cont& wc = w->getChildren();
    for(Window::window_cont::iterator p=wc.begin(); p!=wc.end(); ++p) {
      gAdd(*p);
    }
  }


  struct greater_draw_priority
  {
    bool operator()(const Window *l, const Window *r) {
      return l->getDrawPriority() < r->getDrawPriority();
    }
  };

  void gDraw()
  {
    std::stable_sort(g_draw.begin(), g_draw.end(), greater_draw_priority());
    for(draw_queue::iterator p=g_draw.begin(); p!=g_draw.end(); ++p) {
      MatrixSaver pm(GL_PROJECTION_MATRIX);
      MatrixSaver mm(GL_MODELVIEW_MATRIX);
      (*p)->fitViewport();
      (*p)->draw();
    }
  }

  void gClear()
  {
    g_draw.clear();
  }
}

void View::draw()
{
  setViewport(getPosition(), getSize());
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,getSize().getWidth(), getSize().getHeight(),0, -1,5000);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  const Color& col = getBackgroundColor();
  glClearColor(col.r, col.g, col.b, 0.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  for(window_cont::const_iterator p=m_children.begin(); p!=m_children.end(); ++p) {
    gAdd(*p);
  }
  gDraw();
  gClear();

  SDL_GL_SwapBuffers();
}

void View::setViewport(const Point& pos, const Size& size) const
{
  glViewport(
    GLint(pos.getX()/m_width_scale), GLint(pos.getY()/m_height_scale),
    GLsizei(size.getWidth()/m_width_scale), GLsizei(size.getHeight()/m_height_scale));
}

Point View::toOpenGLCoord(const Point& point) const
{
  return Point(point.getX()*m_width_scale, point.getY()*m_height_scale);
}

Point View::toWindowCoord(const Point& point) const
{
  return Point(point.getX()/m_width_scale, point.getY()/m_height_scale);
}

const Size& View::getWindowSize() const
{
  return m_window_size;
}







/*
  App
*/

App* App::instance() { return m_instance; }

void handleSDLEvent(const SDL_Event& evt)
{
  App *app = App::instance();
  switch(evt.type) {
  case SDL_VIDEORESIZE:
    {
      app->getEventServer().queueEvent(new ResizeEvent(
        EVT_APP_RESIZE, 0, 0, Size(evt.resize.w, evt.resize.h), Size()));
    }
    break;

  case SDL_KEYDOWN:
    app->getEventServer().queueEvent(new KeyboardEvent(
      EVT_KEYDOWN, 0, app->getKeyFocus(),
      evt.key.keysym.scancode, evt.key.keysym.mod,
      evt.key.keysym.sym, evt.key.keysym.unicode));
    break;
  case SDL_KEYUP:
    app->getEventServer().queueEvent(new KeyboardEvent(
      EVT_KEYUP, 0, app->getKeyFocus(),
      evt.key.keysym.scancode, evt.key.keysym.mod,
      evt.key.keysym.sym, evt.key.keysym.unicode));
    break;

  case SDL_MOUSEBUTTONDOWN:
    app->getEventServer().queueEvent(new MouseEvent(
      EVT_MOUSE_BUTTONDOWN, 0, app->getMouseFocus(),
      Point(evt.button.x, evt.button.y), Point(),
      MouseEvent::TranslateSDLButton(evt.button.button)));
    break;
  case SDL_MOUSEBUTTONUP:
    app->getEventServer().queueEvent(new MouseEvent(
      EVT_MOUSE_BUTTONUP, 0, app->getMouseFocus(),
      Point(evt.button.x, evt.button.y), Point(),
      MouseEvent::TranslateSDLButton(evt.button.button)));
    break;
  case SDL_MOUSEMOTION:
    app->getEventServer().queueEvent(new MouseEvent(
      EVT_MOUSE_MOVE, 0, app->getMouseFocus(),
      Point(evt.motion.x, evt.motion.y), Point(evt.motion.xrel, evt.motion.yrel),
      MouseEvent::TranslateSDLButtonState(evt.motion.state)));
    break;

  case SDL_JOYBUTTONDOWN:
    app->getEventServer().queueEvent(new JoyEvent(
      EVT_JOY_BUTTONDOWN, 0, app->getKeyFocus(),
      evt.jbutton.which, evt.jbutton.button,
      0, 0));
    break;
  case SDL_JOYBUTTONUP:
    app->getEventServer().queueEvent(new JoyEvent(
      EVT_JOY_BUTTONUP, 0, app->getKeyFocus(),
      evt.jbutton.which, evt.jbutton.button,
      0, 0));
    break;
  case SDL_JOYAXISMOTION:
    app->getEventServer().queueEvent(new JoyEvent(
      EVT_JOY_AXIS, 0, app->getKeyFocus(),
      evt.jaxis.which, 0,
      (evt.jaxis.axis==0 ? evt.jaxis.value : 0), (evt.jaxis.axis==1 ? evt.jaxis.value : 0)));
    break;

  case SDL_QUIT:
    app->getEventServer().queueEvent(new Event(EVT_APP_EXIT, 0, 0));
    break;
  }
};

void App::waitSDLEvent()
{
  SDL_Event e;
  while(!m_end_flag && SDL_PollEvent(&e)) {
    handleSDLEvent(e);
    if(e.type==SDL_QUIT) {
      m_end_flag = true;
    }
  }
}

App* App::m_instance = 0;

App::App(int argc, char **argv) :
  m_argc(argc),
  m_argv(argv),
  m_default_font_size(12),
  m_default_font(0),
  m_end_flag(false),
  m_view(0),
  m_key_focus_window(0),
  m_mouse_focus_window(0)
{
  assert(m_instance==0);
  m_instance = this;

  if(SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK)<0) {
    throw std::runtime_error(SDL_GetError());
  }

  SDL_EnableUNICODE(1);
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  setDefaultForegroundColor(Color());
  setDefaultBackgroundColor(Color(0.0f, 0.0f, 0.0f, 0.7f));
  setDefaultBorderColor(Color());

#ifdef _WIN32
  char path[256];
  if(GetWindowsDirectoryA(path, 256)) {
    setDefaultFont(string(path)+"\\Fonts\\msgothic.ttc");
  }
#else
  setDefaultFont("/usr/share/fonts/truetype/ipamona/ipag-mona.ttf");
#endif
}

App::~App()
{
  for(font_cont::iterator p=m_fonts.begin(); p!=m_fonts.end(); ++p) {
    delete p->second;
  }
  m_fonts.clear();

  View::instance()->release();

  SDL_Quit();
}

int App::exec()
{
  loop();
  release();
  return 0;
}

void App::stop()
{
  m_end_flag = true;
}

void App::loop()
{
  while(!m_end_flag) {
    update();
    draw();
    SDL_Delay(1);
  }
}

void App::update()
{
  waitSDLEvent();
  getEventServer().deliverEvent();
  View::instance()->update();
}

bool App::handleEvent(const Event& evt)
{
  if(evt.getType()==EVT_APP_DESTROY_WINDOW) {
    if(Window *w = evt.getSrc()) {
      w->release();
    }
    return true;
  }
  else if(evt.getType()==EVT_DD_DROP) {
    if(!evt.isHandled()) {
      if(DDItem *dd = dynamic_cast<DDItem*>(evt.getSrc())) {
        dd->setPosition(dd->getDragBeginPosition());
      }
    }
  }

  return false;
}


void App::draw()
{
  View::instance()->draw();
}

EventServer& App::getEventServer() { return m_evtserver; }
const Color& App::getDefaultForegroundColor() const { return m_fg_color; }
const Color& App::getDefaultBackgroundColor() const { return m_bg_color; }
const Color& App::getDefaultBorderColor() const { return m_bd_color; }
Font* App::getDefaultFont() const { return m_default_font; }
Font* App::getFont(const string& path, size_t size) const
{
  char name[128];
  sprintf(name, "%s_%d", path.c_str(), size);
  font_cont::const_iterator i = m_fonts.find(name);
  if(i!=m_fonts.end()) {
    return i->second;
  }
  return 0;
}


Window* App::getKeyFocus() const { return m_key_focus_window; }
Window* App::getMouseFocus() const { return m_mouse_focus_window; }
View* App::getView() { return m_view; }


void App::initOpenGLResource()
{
  loadFont(m_default_font_name, m_default_font_size);
}

bool App::loadFont(const string& path, size_t size)
{
  char name[128];
  sprintf(name, "%s_%d", path.c_str(), size);
  if(!!m_fonts[name]) {
    return true;
  }

  Font *font = new FTGLTextureFont(path.c_str());
  if(font->Error()==0) {
    font->FaceSize(size);
    m_fonts[path] = font;
    if(!m_default_font) {
      m_default_font = font;
    }
    return true;
  }
  else {
    delete font;
  }

  return false;
}

void App::releaseFont(const string& path)
{
  delete m_fonts[path];
  m_fonts[path] = 0;
}

void App::setDefaultForegroundColor(const Color& color) { m_fg_color = color; }
void App::setDefaultBackgroundColor(const Color& color) { m_bg_color = color; }
void App::setDefaultBorderColor(const Color& color) { m_bd_color = color; }
void App::setDefaultFont(Font *font) { if(!!font) m_default_font = font; }
void App::setDefaultFont(const string& name) { m_default_font_name = name; }
void App::setDefaultFontSize(size_t v) { m_default_font_size=v; }

void App::setKeyFocus(Window *window)
{
  if(m_key_focus_window!=window) {
    if(m_key_focus_window) {
      getEventServer().queueEvent(new Event(EVT_LOSINGKEYFOCUS, 0, m_key_focus_window));
    }
    m_key_focus_window = window;
    getEventServer().queueEvent(new Event(EVT_GAININGKEYFOCUS, 0, m_key_focus_window));
  }
}

void App::setMouseFocus(Window *window)
{
  if(m_mouse_focus_window!=window) {
    if(m_mouse_focus_window) {
      getEventServer().queueEvent(new Event(EVT_LOSINGMOUSEFOCUS, 0, m_mouse_focus_window));
    }
    m_mouse_focus_window = window;
    getEventServer().queueEvent(new Event(EVT_GAININGMOUSEFOCUS, 0, m_mouse_focus_window));
  }
}

void App::setView(View *view)
{
  m_view = view;
}

} // sgui
