#include <vector>
#include <deque>

#include "vector.h"
#include "shader_object.h"
#include "frame_buffer_object.h"

#include "sgui.h"


typedef boost::intrusive_ptr<FrameBufferObject> fbo_ptr;
typedef boost::intrusive_ptr<ProgramObject> po_ptr;


float GetRandom()
{
  return float(::rand()%1000)/1000.0f;
}


class ScreenMatrix
{
private:
  float m_modelview_matrix[16];
  float m_projection_matrix[16];

public:
  ScreenMatrix(int width=640, int height=480)
  {
    glGetFloatv(GL_MODELVIEW_MATRIX, m_modelview_matrix);
    glGetFloatv(GL_PROJECTION_MATRIX, m_projection_matrix);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, width, height, 0, -1.0f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    vector4 pos = vector4(0.0f, 0.0f, 500.0f);
    vector4 tar = vector4(0.0f, 0.0f, 0.0f);
    vector3 dir = vector3(0.0f, 1.0f, 0.0f);
    gluLookAt(
      pos.x, pos.y, pos.z,
      tar.x, tar.y, tar.z,
      dir.x, dir.y, dir.z);
  }
  
  ~ScreenMatrix()
  {
    int mode;
    glGetIntegerv(GL_MATRIX_MODE, &mode);

    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(m_projection_matrix);
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(m_modelview_matrix);

    glMatrixMode(mode);
  }
};

void DrawRect(const vector2& ur, const vector2& bl)
{
  ScreenMatrix sm(640, 480);

  glBegin(GL_QUADS);
  glTexCoord2f(0.0f, 1.0f);
  glVertex2f(bl.x, bl.y);
  glTexCoord2f(0.0f, 0.0f);
  glVertex2f(bl.x, ur.y);
  glTexCoord2f(1.0f, 0.0f);
  glVertex2f(ur.x, ur.y);
  glTexCoord2f(1.0f, 1.0f);
  glVertex2f(ur.x, bl.y);
  glEnd();
}

void DrawTriangle(const vector2& ur, const vector2& bl)
{
  ScreenMatrix sm(640, 480);

  glBegin(GL_TRIANGLES);
  glTexCoord2f(0.0f, 2.0f);
  glVertex2f(bl.x, bl.y-ur.y);
  glTexCoord2f(0.0f, 0.0f);
  glVertex2f(bl.x, ur.y);
  glTexCoord2f(2.0f, 0.0f);
  glVertex2f(ur.x*2.0, ur.y);
  glEnd();
}


class Line : public Object
{
private:
  std::deque<vector2> m_vertex;
  std::vector<vector4> m_color;
  vector2 m_pos;
  vector2 m_direction;
  size_t m_frame;

public:
  Line(const size_t length=100) : m_frame(0)
  {
    m_direction = vector2(0.0f, 5.0f);
    m_pos = vector2(0, 0);
    m_vertex.resize(length, m_pos);

    float u = 1.0f/length;
    for(int i=length; i>=0; --i) {
      m_color.push_back(vector4(u*i, u*i, 1.0f, u*i));
    }

  };

  void update()
  {
    const float speed = 1.5f;

    m_pos+=m_direction;
    m_vertex.push_front(m_pos);
    m_vertex.pop_back();

    if(++m_frame==300) {
      m_frame = 0;
      m_pos = vector2(0, 0);
      std::fill(m_vertex.begin(), m_vertex.end(), m_pos);
    }

    if(GetRandom()<0.05f) {
      float f = GetRandom();
      if(f < 0.5f) {
        m_direction = vector2(speed, 0.0f);
      }
      else {
        m_direction = vector2(0.0f, speed);
      }
    }
  }

  void draw()
  {
    glBegin(GL_LINE_STRIP);
    for(int i=0; i<m_vertex.size(); ++i) {
      glColor4fv(m_color[i].v);
      glVertex2fv(m_vertex[i].v);
    }
    glEnd();
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
  }
};


class LineTexture : public Object
{
private:
  typedef boost::intrusive_ptr<Line> line_ptr;
  typedef std::vector<line_ptr> line_cont;

  fbo_ptr m_fbo;
  line_cont m_lines;

public:
  LineTexture(size_t width=128, size_t height=128)
  {
    m_fbo = new FrameBufferObject(width, height);
    for(size_t i=0; i<5; ++i) {
      m_lines.push_back(new Line());
    }
  }

  void update()
  {
    for(line_cont::iterator p=m_lines.begin(); p!=m_lines.end(); ++p) {
      (*p)->update();
    }
  }

  void draw()
  {
    ScreenMatrix sm(m_fbo->getWidth(), m_fbo->getHeight());

    m_fbo->enable();
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glLineWidth(2.0f);
    for(line_cont::iterator p=m_lines.begin(); p!=m_lines.end(); ++p) {
      (*p)->draw();
    }
    glLineWidth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);

    m_fbo->disable();
  }

  void assign() { m_fbo->assign(); }
  void disassign() { m_fbo->disassign(); }
  fbo_ptr getFBO() { return m_fbo; }
};



class Box : public Object
{
private:
  std::vector<vector3> m_vertex;
  std::vector<vector3> m_normal;
  std::vector<vector2> m_tex;

public:
  Box(const vector4& ur, const vector4& bl)
  {
    vector3 n;

    n = vector3(1.0f, 0.0f, 0.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_vertex.push_back(vector3(ur.x, ur.y, bl.z));
    m_vertex.push_back(vector3(ur.x, ur.y, ur.z));
    m_vertex.push_back(vector3(ur.x, bl.y, ur.z));
    m_vertex.push_back(vector3(ur.x, bl.y, bl.z));

    n = vector3(-1.0f, 0.0f, 0.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_vertex.push_back(vector3(bl.x, ur.y, ur.z));
    m_vertex.push_back(vector3(bl.x, ur.y, bl.z));
    m_vertex.push_back(vector3(bl.x, bl.y, bl.z));
    m_vertex.push_back(vector3(bl.x, bl.y, ur.z));

    n = vector3(0.0f, 1.0f, 0.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_vertex.push_back(vector3(ur.x, ur.y, ur.z));
    m_vertex.push_back(vector3(ur.x, ur.y, bl.z));
    m_vertex.push_back(vector3(bl.x, ur.y, bl.z));
    m_vertex.push_back(vector3(bl.x, ur.y, ur.z));

    n = vector3(0.0f, -1.0f, 0.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_vertex.push_back(vector3(ur.x, bl.y, bl.z));
    m_vertex.push_back(vector3(ur.x, bl.y, ur.z));
    m_vertex.push_back(vector3(bl.x, bl.y, ur.z));
    m_vertex.push_back(vector3(bl.x, bl.y, bl.z));

    n = vector3(0.0f, 0.0f, 1.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_vertex.push_back(vector3(ur.x, bl.y, ur.z));
    m_vertex.push_back(vector3(ur.x, ur.y, ur.z));
    m_vertex.push_back(vector3(bl.x, ur.y, ur.z));
    m_vertex.push_back(vector3(bl.x, bl.y, ur.z));

    n = vector3(0.0f, 0.0f, -1.0f);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_normal.push_back(n);
    m_tex.push_back(vector2(1.0f, 1.0f));
    m_tex.push_back(vector2(1.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 0.0f));
    m_tex.push_back(vector2(0.0f, 1.0f));
    m_vertex.push_back(vector3(ur.x, ur.y, bl.z));
    m_vertex.push_back(vector3(ur.x, bl.y, bl.z));
    m_vertex.push_back(vector3(bl.x, bl.y, bl.z));
    m_vertex.push_back(vector3(bl.x, ur.y, bl.z));
  }

  void draw()
  {
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    glVertexPointer(3, GL_FLOAT, 0, m_vertex.front().v);
    glNormalPointer(GL_FLOAT, 0, m_normal.front().v);
    glTexCoordPointer(2, GL_FLOAT, 0, m_tex.front().v);

    glDrawArrays(GL_QUADS, 0, 4*6);

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
  }
};




enum {
  BU_FBO1 = 1,
  BU_FBO2,
  BU_FBO3,
};

class FBODialog : public sgui::Dialog
{
typedef sgui::Dialog Super;
private:
  fbo_ptr m_fbo;

public:
  FBODialog(const sgui::wstring& title, fbo_ptr fbo) :
    Super(sgui::View::instance(), sgui::Point(5,5), sgui::Size(fbo->getWidth(), fbo->getHeight()+25), title),
    m_fbo(fbo)
  {
  }

  void draw()
  {
    Super::draw();

    // _CAȌ㕔25 
    glTranslatef(0, 25, 0);

    glEnable(GL_TEXTURE_2D);
    m_fbo->assign();
    sgui::DrawRect(sgui::Rect(sgui::Size(m_fbo->getWidth(), m_fbo->getHeight())));
    m_fbo->disassign();
    glDisable(GL_TEXTURE_2D);
  }
};

class Background : public sgui::Panel
{
typedef sgui::Panel Super;
private:
  typedef boost::intrusive_ptr<Box> box_ptr;
  typedef boost::intrusive_ptr<LineTexture> tex_ptr;

  fbo_ptr m_fbo_line;
  fbo_ptr m_fbo_glow;
  fbo_ptr m_fbo_glow2;
  po_ptr m_glow;
  box_ptr m_box;
  tex_ptr m_tex;

  bool m_end_flag;
  size_t m_width;
  size_t m_height;

public:
  Background(sgui::Window *parent) :
    Super(parent, sgui::Point(), parent->getSize())
  {
    listenEvent(sgui::EVT_BUTTON_UP);
    new sgui::Button(this, sgui::Point(5, 5), sgui::Size(70, 16), L"fbo1", BU_FBO1);
    new sgui::Button(this, sgui::Point(5, 25), sgui::Size(70, 16), L"fbo2", BU_FBO2);
    new sgui::Button(this, sgui::Point(5, 45), sgui::Size(70, 16), L"fbo3", BU_FBO3);

    m_fbo_line = new FrameBufferObject(640, 480);
    m_fbo_glow = new FrameBufferObject(320, 240);
    m_fbo_glow2 = new FrameBufferObject(320, 240);
    m_glow = new ProgramObject();
    m_glow->attach(new FragmentShader("glow.fsh"));
    m_glow->link();

    m_box = new Box(vector4(50.0f), vector4(-50.0f));
    m_tex = new LineTexture();
  }

  void draw()
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, 1.333f, 1.0f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    vector4 pos = vector4(0.0f, 100.0f, 200.0f);
    vector4 tar = vector4(0.0f, 0.0f, 0.0f);
    vector3 dir = vector3(0.0f, 1.0f, 0.0f);
    gluLookAt(
      pos.x, pos.y, pos.z,
      tar.x, tar.y, tar.z,
      dir.x, dir.y, dir.z);

    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, vector4(1000.0f, 1000.0f, 500.0f).v);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  vector4(1.0f, 1.0f, 1.0f).v);

    glClearColor(0.2f, 0.2f, 0.2f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    static float rot = 0.0f;
    glPushMatrix();
    glRotatef(rot, 0.0f, 1.0f, 0.0f);
    rot+=0.5f;

    m_tex->draw();

    // ܂͕ʂɕ` 
    m_box->draw();

    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);

    // eNX`Ԃfboɕ` 
    m_fbo_line->enable();
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    m_tex->assign();
    m_box->draw();
    m_tex->disassign();
    m_fbo_line->disable();
    glDisable(GL_DEPTH_TEST);

    // eNX`ԂZ 
    m_fbo_line->assign();
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    DrawRect(vector2(640, 480), vector2(0, 0));
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    m_fbo_line->disassign();


    // ڂpkobt@ɃRs[ 
    m_fbo_glow->enable();
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    m_glow->enable();
    m_glow->setUniform1f("width", float(m_fbo_glow->getWidth()));
    m_glow->setUniform1f("height", float(m_fbo_glow->getHeight()));
    m_glow->setUniform1i("pass", 1);
    m_fbo_line->assign();
    DrawRect(vector2(640, 480), vector2(0, 0));
    m_glow->disable();
    m_fbo_glow->disable();

    m_fbo_glow2->enable();
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    m_glow->enable();
    m_glow->setUniform1f("width", float(m_fbo_glow2->getWidth()));
    m_glow->setUniform1f("height", float(m_fbo_glow2->getHeight()));
    m_glow->setUniform1i("pass", 2);
    m_fbo_glow->assign();
    DrawRect(vector2(640, 480), vector2(0, 0));
    m_glow->disable();
    m_fbo_glow2->disable();


    // ڂ̂Z 
    m_fbo_glow2->assign();
    glColor3f(0.3f, 0.3f, 1.0f);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    DrawRect(vector2(640, 480), vector2(0, 0));
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor3f(1.0f, 1.0f, 1.0f);

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glPopMatrix();
  }


  void update()
  {
    m_tex->update();
  }



  bool handleEvent(const sgui::Event& evt)
  {
    if(evt.getType()==sgui::EVT_BUTTON_UP) {
      int id = evt.getSrc() ? evt.getSrc()->getID() : 0;
      if(id==BU_FBO1) {
        new FBODialog(L"texture", m_tex->getFBO());
        return true;
      }
      else if(id==BU_FBO2) {
        new FBODialog(L"glow1", m_fbo_glow);
        return true;
      }
      else if(id==BU_FBO3) {
        new FBODialog(L"glow2", m_fbo_glow2);
        return true;
      }
    }
    return Super::handleEvent(evt);
  }
};




class App : public sgui::App
{
public:
  App(int argc, char **argv) : sgui::App(argc, argv)
  {
    setView(new sgui::View("FBO test", sgui::Size(640, 480), sgui::Size(640, 480)));
    glewInit();

    if(!GLEW_VERSION_2_0) {
      new sgui::MessageDialog(L"error", L"OpenGL2.0 Ή̂߁AVF[_gpł܂B");
    }
    else if(!GLEW_EXT_framebuffer_object) {
      new sgui::MessageDialog(L"error", L"EXT_framebuffer_object Ή̂߁AVF[_gpł܂B");
    }
    else {
      new Background(getView());
    }
  }

  void loop()
  {
    const size_t interval = 50;
    Uint32 cur_tick = SDL_GetTicks()*3;
    Uint32 pre_frame = 0;
    Uint32 count = 0;
    while(!m_end_flag) {
      cur_tick = SDL_GetTicks()*3;
      if(cur_tick-pre_frame < interval) {
        SDL_Delay(std::max<int>(int(pre_frame+interval - cur_tick)/3, 0));
        pre_frame+=interval;
      }
      else {
        pre_frame = cur_tick;
      }

      update();
      draw();
    }
  }

  bool handleEvent(const sgui::Event& e) {
    if(e.getType()==sgui::EVT_KEYUP) {
      const sgui::KeyboardEvent& k = dynamic_cast<const sgui::KeyboardEvent&>(e);
      if(k.getKey()==SDLK_ESCAPE) {
        stop();
        return true;
      }
    }
    return sgui::App::handleEvent(e);
  }
};



int main(int argc, char *argv[])
{
  return (new App(argc, argv))->exec();
}
